library bloc_stream;

import 'dart:async';

import 'package:meta/meta.dart';

typedef Action<B extends BlocStream<State>, State> = FutureOr<void> Function(
    B bloc, void Function(State) add);

abstract class BlocStream<State> extends Stream<State> {
  BlocStream(State initialValue) {
    _value = initialValueOverride ?? initialValue;
    _bindActionsToState();
  }

  State? get initialValueOverride => null;
  late State _value;
  State get value => _value;

  final _actionController = StreamController<dynamic>(sync: true);
  final _stateController = StreamController<State>.broadcast(sync: true);

  @override
  bool get isBroadcast => true;

  @override
  StreamSubscription<State> listen(
    void Function(State)? onData, {
    Function? onError,
    void Function()? onDone,
    bool? cancelOnError,
  }) =>
      _stateController.stream.listen(
        onData,
        onError: onError,
        onDone: onDone,
        cancelOnError: cancelOnError,
      );

  /// Add actions to be processed. Implements Sink
  void add<B extends BlocStream<State>>(Action<B, State> action) {
    if (_actionController.isClosed) return;
    _actionController.add(action as dynamic);
  }

  @mustCallSuper
  void willUpdateValue(State current, State next) {}

  void _bindActionsToState() {
    _actionController.stream.asyncMap(_processAction).listen(
          null,
          onError: (err) => print('[$runtimeType]: Error in action: $err'),
          cancelOnError: false,
          onDone: _stateController.close,
        );
  }

  FutureOr<void> _processAction(dynamic action) =>
      action(this, _handleNewState);

  void _handleNewState(State newState) {
    if (newState == value) return;
    willUpdateValue(value, newState);
    _value = newState;
    _stateController.add(newState);
  }

  /// Helper method to reduce boilerplate when composing BLoC's together.
  final _closers = <void Function()>[];
  void cancelOnClose(StreamSubscription subscription,
      [void Function()? onClose]) {
    _closers.add(() {
      subscription.cancel();
      if (onClose != null) onClose();
    });
  }

  @mustCallSuper
  Future<void> close() {
    _closers.forEach((fn) => fn());
    return _actionController.close();
  }
}
